# Variables for inspecting event dictionaries


PORTAL_RECORD_PK_KEY = 'pgh_obj_id'
OPERATION_KEY = 'operation'
OPERATION_CONFIRM_SYNCED = 'confirm_synced'
SYNCMODE_KEY = 'syncmode'
SYNC_COMPLETED_KEY = 'synced'



import logging
from typing import Any, TYPE_CHECKING
from uuid import UUID
from django.conf import settings
from django.db.models import Manager, Q, QuerySet
from django.contrib.contenttypes.models import ContentTypeManager, ContentType
from django.db import DatabaseError
from django.utils.module_loading import import_string
from pghistory.models import EventQuerySet, EventsQuerySet

if TYPE_CHECKING:
    from .models import ConcordEvent, SyncableModel
    from concordia_tests.latchstring.models import User

from console_base.managers import LCBaseQuerySet
from .constants import PORTAL_RECORD_PK_KEY, SYNC_COMPLETED_KEY

logger = logging.getLogger(__name__)
try:
    SYNC_TABLES_Q = import_string(settings.CONCORD_SYNC_TABLES_QS)
except (AttributeError, ImportError):
    print('settings does not define CONCORD_SYNC_TABLES_QS')
    SYNC_TABLES_Q = Q()


# ---------------------------------------------------------------------------
def set_contenttype_kwargs(**kwargs) -> dict[str, Any]:  # type: ignore[no-untyped-def]
    """Lookup Generic values from schedule keyword if it was provided."""
    record = kwargs.pop('record', None)
    record_data: dict[str, Any] = kwargs.pop('record_data') or {}

    if record:
        try:
            kwargs['object_id'] = record_data.get(PORTAL_RECORD_PK_KEY, '') or record.id
            kwargs['table'] = ContentType.objects.get_for_model(record)
        except AttributeError:
            raise ValueError("'record' value must be a Model Object!")

    return kwargs


# ---------------------------------------------------------------------------
class ConcordiaContentTypeManager(ContentTypeManager):
    def get_queryset(self):
        if SYNC_TABLES_Q:
            return super().get_queryset().filter(SYNC_TABLES_Q).order_by('app_label')
        return super().get_queryset().order_by('app_label')


# ---------------------------------------------------------------------------
class BaseCanonicalIdTrackManager(Manager.from_queryset(LCBaseQuerySet)):  # type: ignore
    def create(self, **kwargs):
        return super().create(**set_contenttype_kwargs(**kwargs))

    def get_or_create(self, defaults=None, **kwargs):
        if defaults:
            defaults = set_contenttype_kwargs(**defaults)
        return super().get_or_create(defaults=defaults, **set_contenttype_kwargs(**kwargs))

    def update_or_create(self, defaults=None, **kwargs):
        if defaults:
            defaults = set_contenttype_kwargs(**defaults)
        return super().update_or_create(defaults=defaults, **set_contenttype_kwargs(**kwargs))


# ---------------------------------------------------------------------------
class CanonicalIdDeleteLogManager(BaseCanonicalIdTrackManager):
    model: 'SyncableModel'

    def record_deletion(
        self,
        record: 'SyncableModel',
        data: dict | None = None,
    ) -> QuerySet['SyncableModel'] | None:
        """
        Record a Canonical ID deletion instance.
        """
        try:
            delete = self.model.objects.create(
                record=record,
                record_data=data,
                name=data.get('name', 'N/A') if isinstance(data, dict) else str(record),
                cid=data['cid'] if isinstance(data, dict) else record.cid,
            )
            return delete
        except DatabaseError as e:
            logger.exception('Unable to record Canonical ID deletion: %e', e)
            pass


# ---------------------------------------------------------------------------
class CanonicalIdSwapManager(BaseCanonicalIdTrackManager):
    """Manager for CanonicalIdSwap model."""

    model: 'SyncableModel'

    def record_swap(
        self,
        record: 'SyncableModel',
        old: UUID,
        data: dict | None = None,
    ) -> QuerySet['SyncableModel'] | None:
        """
        Record a Canonical ID Swap instance.
        """
        try:
            swap = self.model.objects.create(
                record=record,
                record_data=data,
                name=data.get('name', 'N/A') if isinstance(data, dict) else str(record),
                cid=data['cid'] if isinstance(data, dict) else record.cid,
                old=old,
            )
            swap.apply()
            return swap
        except DatabaseError as e:
            logger.exception('Unable to record Canonical ID swap: %e', e)
            pass


# ---------------------------------------------------------------------------
class CanonicalIdMapManager(BaseCanonicalIdTrackManager):
    """Manager for CanonicalIdMap model."""

    model: 'SyncableModel'

    def record_map(
        self,
        record: 'SyncableModel',
        old: UUID,
        data: dict | None = None,
    ) -> QuerySet['SyncableModel'] | None:
        """
        Record a Canonical ID Map instance.
        """
        try:
            mapping = self.model.objects.create(
                record=record,
                record_data=data,
                name=data.get('name', 'N/A') if isinstance(data, dict) else str(record),
                cid=data['cid'] if isinstance(data, dict) else record.cid,
                old=old,
            )
            mapping.apply()
            return mapping
        except DatabaseError as e:
            logger.exception('Unable to record Canonical ID map: %e', e)
            pass


# ---------------------------------------------------------------------------
class SyncedEventMixin:
    def synced(self) -> QuerySet['ConcordEvent']:
        """
        Get all Console Events that have been synced with Publisher.
        """
        qs = self._clone()  # type: ignore[attr-defined]
        return qs.filter((f'pgh_context__{SYNC_COMPLETED_KEY}', True))

    def unsynced(self) -> QuerySet['ConcordEvent']:
        """
        Get all Console Events that have not been synced with Publisher.
        """
        qs = self._clone()  # type: ignore[attr-defined]

        return qs.filter(
            Q(pgh_context__has_key=SYNC_COMPLETED_KEY, _negated=True)
            | Q((f'pgh_context__{SYNC_COMPLETED_KEY}', False))
        )


# ---------------------------------------------------------------------------
class ConsoleEventQuerySet(SyncedEventMixin, EventQuerySet):
    pass


# ---------------------------------------------------------------------------
class ConcordEventsQuerySet(SyncedEventMixin, EventsQuerySet):
    pass


# ---------------------------------------------------------------------------
class SyncQuerySet(LCBaseQuerySet):
    """
    Used for type hints
    """

    def tenant(self, user: 'User') -> QuerySet:
        return self.all()



from ipware.ip import get_client_ip
import logging
import pghistory
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest

if TYPE_CHECKING:
    from concordia_tests.latchstring.models import User

from .choices import ORIGIN_SERVER

logger = logging.getLogger(__name__)


class ConcordiaWSGIRequest(WSGIRequest):
    """
    Nearly exact copy of `pghistory.middleware.WSGIRequest` to set
    the context fields we're interested in.

    Although Django's auth middleware sets the user in middleware,
    apps like django-rest-framework set the user in the view layer.
    This creates issues for pghistory tracking since the context needs
    to be set before DB operations happen.

    This special WSGIRequest updates pghistory context when
    the `request.user` attribute is updated.
    """

    def __setattr__(self, attr, value):
        if attr == 'user' and value and value.is_authenticated:
            pghistory.context(**user_context(value))

        return super().__setattr__(attr, value)


def ConcordiaHistoryMiddleware(get_response):  # noqa
    """
    Add User details to History context.
    """

    def middleware(request):
        if (
            request.method in settings.PGHISTORY_MIDDLEWARE_METHODS
            and request.path not in settings.SKIP_MIDDLEWARE_URLS
        ):
            user = getattr(request, 'user', None)
            if user and user.is_authenticated:
                context = user_context(user)
            else:
                context = {'user': None}

            ip, _ = get_client_ip(request)

            # Console Sync activity will be local, and we're already
            # adding the appliance source, so skip the IP to save
            # database space.
            if ip != '127.0.0.1':
                context['ip'] = ip

            # The `context` must be initialized in a context manager!
            # So, context must always be initialized here, even if the
            # user has not authenticated, so the `pghistory.context`
            # function call in ConcordiaWSGIRequest can update pre-existing
            # context object.
            with pghistory.context(**context):
                if isinstance(request, WSGIRequest):
                    request.__class__ = ConcordiaWSGIRequest

                return get_response(request)

        return get_response(request)

    return middleware


def user_context(user: 'User') -> dict:
    """
    Return user context dictionary
    """
    try:
        return {
            'user': user.pk,
            'user_cid': user.cid,
            'username': user.username,
            ORIGIN_SERVER.CID: settings.CONCORD_MQ_TENANT,
            ORIGIN_SERVER.Name: settings.SYSTEM_NAME,
        }
    except AttributeError:
        logger.error('Unable to set pghistory.context for User: %s', user)
        return {}



# Tasks for syncing via Concordia MQ
from celery import shared_task
from django.conf import settings
from lc_celery_beat.task_classes import LCSingleton

from ..channels import ROOM_CHANNEL_REGISTRY
from .publishing import schedule_publish_local_events_to_concord_room


@shared_task(base=LCSingleton, lock_expiry=43200, raise_on_duplicate=True)
def confirm_events_synced():
    """
    The TriggerChannel listener may not have been running, a task may
    have failed, so some Events may not be marked as synchronized.

    This task is to ensure that all ConcordEvent model records will
    be synchronized. It should be run on a recurring basis.
    """
    if settings.SYNC_PUBLISHER:
        return

    for channel_name, trigger_channel in ROOM_CHANNEL_REGISTRY.items():
        EventModel = trigger_channel.model
        for record in EventModel.objects.unsynced().distinct('phg_obj_id'):
            schedule_publish_local_events_to_concord_room(
                channel=trigger_channel,
                data=record.pgh_data,
            )


__all__ = [
    'confirm_events_synced',
]



from lchttp.json import json_loads
import logging
from typing import TypedDict
from channels.consumer import AsyncConsumer

logger = logging.getLogger(__name__)


class MessageEvent(TypedDict):
    type: str
    room: str
    body: str


class ClarionConsumer(AsyncConsumer):
    """
    ClarionConsumer to be subclassed by other consumers.
    """

    async def connect(self) -> None:
        """
        Connect method to be subclassed by consumers.
        """
        pass

    async def disconnect(self, event: int) -> None:
        """
        Disconnect method to be subclassed by consumers.
        """
        pass

    async def receive(self, clarion_message: MessageEvent) -> None:
        """
        Receive method to be subclassed by consumers.
        """
        pass

    async def clarion_connect(self, event: int) -> None:
        """
        Initialize connection.
        """
        await self.connect()

    async def clarion_disconnect(self, event: int) -> None:
        """
        Initialize disconnection.
        """
        await self.disconnect(0)

    async def clarion_receive(self, event: MessageEvent) -> None:
        """
        Handle an incoming message from Clarion and
        forward it to Channels.
        """
        try:
            event["body"] = json_loads(event["body"])
        except (KeyError, ValueError):
            pass

        await self.receive(event)

    async def clarion_send(self, room: str, body: str) -> None:
        """
        Handle an incoming message from Channels: forward it to Clarion.
        """
        await self.send({"type": "clarion.send", "room": room, "body": body})



from console_base.api.serializers import (
    LCHyperlinkedIdentityField,
    LCModelSerializer,
)
from concordia.models import CanonicalIdDeleteLog, CanonicalIdMap, CanonicalIdSwap


# ---------------------------------------------------------------------------
class CanonicalIdDeleteLogSerializer(LCModelSerializer):
    url = LCHyperlinkedIdentityField(view_name='api-canonicaliddeletelog-detail')

    class Meta(LCModelSerializer.Meta):
        model = CanonicalIdDeleteLog
        fields = (
            'url',
            'cid',
            'name',
            'table',
            'object_id',
            'app_label',
            'table_name',
        )


# ---------------------------------------------------------------------------
class CanonicalIdMapSerializer(LCModelSerializer):
    url = LCHyperlinkedIdentityField(view_name='api-canonicalidmap-detail')

    class Meta(LCModelSerializer.Meta):
        model = CanonicalIdMap
        fields = (
            'url',
            'cid',
            'local',
            'name',
            'table',
            'object_id',
            'app_label',
            'table_name',
        )


# ---------------------------------------------------------------------------
class CanonicalIdSwapSerializer(LCModelSerializer):
    url = LCHyperlinkedIdentityField(view_name='api-canonicalidswap-detail')

    class Meta(LCModelSerializer.Meta):
        model = CanonicalIdSwap
        fields = (
            'url',
            'cid',
            'old',
            'name',
            'table',
            'object_id',
            'app_label',
            'table_name',
        )



from typing import NamedTuple, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .models import ConcordEvents, SyncableModel
    from concordia_tests.latchstring.models import User


class Record(NamedTuple):
    """
    When a Subscriber posts a message to the Publisher,
    the Subscriber user should have access to the record,
    using the Model.objects.tenant(user) queryset.
    """

    editor: 'User'
    obj: Optional['SyncableModel']
    permitted: bool


class Status(NamedTuple):
    success: bool
    message: dict


class EventData(NamedTuple):
    data: dict
    events: 'ConcordEvents'



# Generated by Django 4.2.10 on 2024-02-27 22:49

import concordia.querysets
import console_base.models.constraints
import console_base.models.fields
import console_base.models.mixins
import console_deps.dirtyfields.dirtyfields
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import sequential_uuids.generators


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("pghistory", "0006_delete_aggregateevent"),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="ConcordEvents",
            fields=[],
            options={
                "verbose_name_plural": "concord events",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("pghistory.events",),
        ),
        migrations.CreateModel(
            name="ConcordiaContentType",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("contenttypes.contenttype",),
            managers=[
                ("objects", concordia.querysets.ConcordiaContentTypeManager()),
            ],
        ),
        migrations.CreateModel(
            name="CanonicalIdSwap",
            fields=[
                ("created", models.DateTimeField(default=django.utils.timezone.now)),
                ("modified", models.DateTimeField(default=django.utils.timezone.now)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "id",
                    console_base.models.fields.UUID6Field(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "cid",
                    console_base.models.fields.CanonicalIdField(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        help_text="Canonical ID assigned to this record, locally or on external systems. Should not be changed, except to reconcile sync errors.",
                    ),
                ),
                ("name", models.TextField()),
                ("object_id", models.TextField(blank=True, null=True)),
                ("old", models.UUIDField()),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "canonical id map",
            },
            bases=(
                console_base.models.mixins.FilterFields,
                console_deps.dirtyfields.dirtyfields.DirtyFieldsMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="CanonicalIdMap",
            fields=[
                ("created", models.DateTimeField(default=django.utils.timezone.now)),
                ("modified", models.DateTimeField(default=django.utils.timezone.now)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "id",
                    console_base.models.fields.UUID6Field(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "cid",
                    console_base.models.fields.CanonicalIdField(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        help_text="Canonical ID assigned to this record, locally or on external systems. Should not be changed, except to reconcile sync errors.",
                    ),
                ),
                ("name", models.TextField()),
                ("object_id", models.TextField(blank=True, null=True)),
                ("local", models.UUIDField()),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "canonical id map",
            },
            bases=(
                console_base.models.mixins.FilterFields,
                console_deps.dirtyfields.dirtyfields.DirtyFieldsMixin,
                models.Model,
            ),
        ),
        migrations.CreateModel(
            name="CanonicalIdDeleteLog",
            fields=[
                ("created", models.DateTimeField(default=django.utils.timezone.now)),
                ("modified", models.DateTimeField(default=django.utils.timezone.now)),
                ("is_active", models.BooleanField(default=True)),
                (
                    "id",
                    console_base.models.fields.UUID6Field(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "cid",
                    console_base.models.fields.CanonicalIdField(
                        default=sequential_uuids.generators.uuid_time_nextval,
                        help_text="Canonical ID assigned to this record, locally or on external systems. Should not be changed, except to reconcile sync errors.",
                    ),
                ),
                ("name", models.TextField()),
                ("object_id", models.TextField(blank=True, null=True)),
                (
                    "table",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "canonical id delete",
            },
            bases=(
                console_base.models.mixins.FilterFields,
                console_deps.dirtyfields.dirtyfields.DirtyFieldsMixin,
                models.Model,
            ),
        ),
        migrations.AddConstraint(
            model_name="canonicalidswap",
            constraint=models.CheckConstraint(
                check=models.Q(("cid", models.F("old")), _negated=True),
                name="old_cid_differs_from_new_cid",
                violation_error_message="Old CID must differ from new CID",
            ),
        ),
        migrations.AddConstraint(
            model_name="canonicalidswap",
            constraint=console_base.models.constraints.LCUniqueConstraint(
                fields=("object_id", "old"),
                name="old_canonical_id_unique",
                violation_error_message="Record swapping this old Canonical ID already exists.",
            ),
        ),
        migrations.AddConstraint(
            model_name="canonicalidmap",
            constraint=models.CheckConstraint(
                check=models.Q(("cid", models.F("local")), _negated=True),
                name="local_cid_differs_from_global_cid",
                violation_error_message="Local CID must differ from global CID",
            ),
        ),
        migrations.AddConstraint(
            model_name="canonicalidmap",
            constraint=console_base.models.constraints.LCUniqueConstraint(
                condition=models.Q(("object_id__isnull", False)),
                fields=("object_id", "local"),
                name="local_canonical_id_unique",
                violation_error_message="Local record mapping to Canonical ID already exists.",
            ),
        ),
        migrations.AddConstraint(
            model_name="canonicaliddeletelog",
            constraint=console_base.models.constraints.LCUniqueConstraint(
                fields=("cid",),
                name="deleted_canonical_id_unique",
                violation_error_message="Delete record for this Canonical ID already exists.",
            ),
        ),
    ]



from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as tr


# ---------------------------------------------------------------------------
class CONTEXT(TextChoices):
    Event = 'pgh_context', tr('Context from local Concord Event')
    Origin = 'origin_context', tr("Subscriber's original Concord Event Context")
    Keys = 'keys', tr("Context Keys from Subscriber's Event Context")


# ---------------------------------------------------------------------------
class SYNCMODES(TextChoices):
    """
    Should records be deleted on the local console only,
    or synchronized among all the consoles?
    """

    Sync = 'sync', tr('Sync Everywhere')
    Local = 'local', tr('Local System Only')


# ---------------------------------------------------------------------------
class CRUD(TextChoices):
    Create = 'create', 'Create'
    Update = 'update', 'Update'
    Delete = 'delete', 'Delete'
    Toggle = 'toggle', 'Toggle'


# ---------------------------------------------------------------------------
class ORIGIN_SERVER(TextChoices):
    CID = 'origin_server_cid', tr('Origin Appliance / Server CID')
    Name = 'origin_server_name', tr('Origin Appliance / Server Name')



from lcconfig import ConcordiaYmlConfig

from .server import ClarionServer

if __name__ == '__main__':
    cfg = ConcordiaYmlConfig().as_typed_dict()

server = ClarionServer(
    server=cfg['server'],
    user=cfg['user'],
    password=cfg['password'],
    application=cfg['asgi_application'],
)

server.run()



import asyncio
from lchttp.json import json_dumps
import logging
from signal import SIGINT, SIGTERM
from typing import Any

from nio import AsyncClient
from nio.events.room_events import RoomMessageText
from nio.exceptions import LocalProtocolError
from nio.responses import JoinError, RoomSendError
from nio.rooms import MatrixRoom

from markdown import markdown

from . import utils

logger = logging.getLogger(__name__)


class ClarionServer:
    """
    Clarion ASGI Server.
    """

    def __init__(
        self,
        server: str,
        user: str,
        password: str,
        application: str,
    ):
        self.server = server
        self.user = user
        self.password = password
        self.application = utils.get_application(application)

        self.client = AsyncClient(self.server, self.user)
        self.client.add_event_callback(self.message_callback, RoomMessageText)

        self.app_scope = {
            "type": "concord",
            "asgi": {
                "version": "3.0",
                "spec_version": "1.0",
            },
        }
        self.queue: asyncio.Queue | None = None
        self.event: asyncio.Event | None = None

    def run(self) -> None:
        """
        Start the synchronous ASGI server.
        """
        logger.info("Starting Clarion ASGI server")
        asyncio.run(self.main())
        logger.info("Clarion ASGI server shutdown complete")

    async def login(self) -> None:
        """
        Login or re-login to the Clarion homeserver.
        """
        await self.client.login(self.password)

    async def main(self, event: asyncio.Event | None = None) -> None:
        """
        Start the asynchronous ASGI server.
        """
        self.queue = asyncio.Queue()
        self.event = event or asyncio.Event()

        if event is None:
            for sig in (SIGINT, SIGTERM):
                asyncio.get_running_loop().add_signal_handler(sig, utils.terminate, self.event, sig)

        await self.login()
        asyncio.create_task(
            self.application(self.app_scope, receive=self.queue.get, send=self.app_send),  # type: ignore[operator]
        )
        asyncio.create_task(
            self.client.sync_forever(timeout=36_000),
        )
        await self.queue.put({
            "type": "clarion.connect",
        })

        logger.info("Clarion ASGI Server startup complete")

        if event is None:
            await self.serve()

    async def serve(self) -> None:
        """
        Server requests forever.
        """
        if self.event is None or self.queue is None:
            return
        await self.event.wait()
        await self.queue.put({
            "type": "clarion.disconnect",
        })
        logger.info("Stopping Clarion ASGI Server")
        await self.client.close()

    async def app_send(self, message: dict) -> None:
        """
        ASGI application `send` method: Dispatch incoming Channel messages.
        """
        logger.debug(f"app_send {message=}")
        await self.clarion_room_send(message["room"], message["body"])

    async def message_callback(self, room: MatrixRoom, event: RoomMessageText) -> None:
        """
        Handle an incoming message from matrix-nio client: put it in the Queue.
        """
        logger.debug("message_callback {room=} {event=}")
        if not self.queue:
            return
        await self.queue.put({
            "type": "clarion.receive",
            "room": room.display_name,
            "user": room.user_name(event.sender),
            "body": event.body,
        })

    async def join_room(self, room_id: str) -> bool:
        """
        Use matrix-nio client to join a room.
        """
        logger.debug(f"Join room {room_id=}")

        for _ in range(10):
            try:
                resp = await self.client.join(room_id)
                if not isinstance(resp, JoinError):
                    return True
                if resp.status_code == "M_UNKNOWN_TOKEN":
                    logger.warning("Reconnecting")
                    await self.login()
                elif resp.status_code in ["M_FORBIDDEN", "M_CONSENT_NOT_GIVEN"]:
                    logger.error("room access is forbidden")
                    return False
                elif resp.status_code == "M_UNKNOWN":
                    logger.error(f"join error: {resp.transport_response.status}")
                    return False
            except LocalProtocolError as e:
                logger.error(f"Send error: {e}")
            logger.warning("Trying again")
        logger.error("Clarion Homeserver not responding")
        return False

    async def send_room_message(self, room_id: str, content: dict[Any, Any]) -> bool:
        """
        Use matrix-nio client to send a Clarion message to a room.
        """
        logger.debug(f"Sending room message in {room_id=}: {content=}")

        for _ in range(10):
            try:
                resp = await self.client.room_send(
                    room_id=room_id,
                    message_type="m.room.message",
                    content=content,
                )
                if not isinstance(resp, RoomSendError):
                    return True
                if resp.status_code == "M_UNKNOWN_TOKEN":
                    logger.warning("Unknown token - reconnecting to Clarion Homeserver")
                    await self.login()
                else:
                    logger.error(f"room send error {resp=}")
                    return False
            except LocalProtocolError as e:
                logger.error(f"Send error: {e}")
            logger.warning("Trying again")
        logger.error("Clarion Homeserver not responding")
        return False

    async def clarion_room_send(self, room: str, message: str) -> None:
        """
        Handle an incoming `clarion.send` message from Channel,
        and forward it to Clarion.
        """
        content = {
            "msgtype": "m.text",
            "body": message,
            "format": "org.matrix.custom.html",
            "formatted_body": markdown(message, extensions=["extra"]),
        }
        await self.send_room_message(room, content)

    async def clarion_record_send(self, room: str, message: str) -> None:
        """
        Handle an incoming `clarion.record` message from Channel,
        and forward it to Clarion.

        The record is dumped to JSON before posting to the room.
        """
        if not isinstance(message, str):
            message = json_dumps(message, default=str, rtype=str)

        content = {
            "msgtype": "m.text",
            "body": message,
        }
        await self.send_room_message(room, content)



from typing import Any

from pghistory.models import Event, Events

from django.conf import settings
from django.db import models
from django.db.models import (
    DO_NOTHING,
    ForeignKey,
)
from django.utils.functional import cached_property
from console_base.choices import CSS
from ..choices import CRUD
from ..constants import OPERATION_KEY
from ..querysets import (
    ConsoleEventQuerySet,
    ConcordEventsQuerySet,
)


# ---------------------------------------------------------------------------
class ConcordEvent(Event):
    """
    Custom event for Console history tracking.
    Subclassed to add any project-wide custom functionality that might be useful.
    """

    pgh_id = models.BigAutoField(primary_key=True)  # type: ignore

    objects = ConsoleEventQuerySet.as_manager()

    class Meta:
        abstract = True


# ---------------------------------------------------------------------------
class ConcordEvents(Events):
    """
    Extend to add User Editor & Appliance Editor
    columns for easier filtering.
    """

    user = ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=DO_NOTHING,
        null=True,
    )

    objects = ConcordEventsQuerySet.as_manager()  # type: ignore[django-manager-missing,assignment,misc]

    class Meta:
        proxy = True
        app_label = 'concordia'
        verbose_name_plural = "concord events"

    @cached_property
    def operation(self):
        """
        Get CRUD operation
        """
        if self.pgh_diff:
            return CRUD.Update

        try:
            return self.pgh_context.metadata.get(OPERATION_KEY, CRUD.Create)
        except AttributeError:
            return CRUD.Create

    @cached_property
    def wizard_action(self) -> dict[str, str]:
        """
        Various wizards are used to keep records current.
        Get name & action if any exist in the context.
        """
        try:
            metadata: dict[str, Any] = self.pgh_context or {}
        except AttributeError:
            return {}

        wizard = metadata.get(settings.WIZARD_NAME)
        if not wizard:
            return {}

        return {
            OPERATION_KEY: metadata.get(settings.WIZARD_ACTION) or '',
            'name': wizard,
        }

    def color_class(self):
        if self.operation == CRUD.Update:
            return CSS.Info
        if self.operation == CRUD.Delete:
            return CSS.Danger
        if self.operation == CRUD.Toggle:
            return CSS.Warning
        return CSS.Success


__all__ = (
    'ConcordEvent',
    'ConcordEvents',
)



from typing import Any, TYPE_CHECKING, Union

from django.apps import apps
from django.utils import timezone

from concordia.events.trail import ConcordEventsTrail
from console_deps.datetime_truncate import truncate_day
from console_base.models import BaseIntegerPKModel, BaseUUIDPKModel
from console_base.views import LCDetailViewBase

if TYPE_CHECKING:
    ActivityStreamBase = LCDetailViewBase
else:
    ActivityStreamBase = object


# ----------------------------------------------------------------------
class ActivityStreamViewerMixin(ActivityStreamBase):
    def update_last_viewed(self) -> None:
        self.request.user.last_viewed_events = timezone.now()
        self.request.user.save(update_fields=['last_viewed_events'])

    def get_context_data(self, **kwargs: dict[Any, Any]) -> dict[Any, Any]:
        user = self.request.user
        events = ConcordEventsTrail(user.last_viewed_events)
        if not events.unviewed_events_exist():
            # use truncate_date to retain the timezone
            events.date = truncate_day(timezone.localtime())

        self.update_last_viewed()

        return super().get_context_data(
            events_since=events.date,
            event_stream=events.unviewed_events(),
            **kwargs,
        )


# ----------------------------------------------------------------------
class ActivityStreamRecordViewerMixin(ActivityStreamBase):
    """
    Get activity stream for entire life cycle of this record
    """

    def get_record(self) -> Union['BaseIntegerPKModel', 'BaseUUIDPKModel']:
        Model = apps.get_model(app_label=self.kwargs['app'], model_name=self.kwargs['model'])
        self.object = Model.objects.filter(cid=self.kwargs['cid']).first()  # noqa
        return self.object

    def get_context_data(self, **kwargs: dict[Any, Any]) -> dict[Any, Any]:
        if record := self.get_record():
            events_since = record.created
            record_css_icon = record.css_icon
            events = ConcordEventsTrail(record.created).record_events(target=record)
        else:
            events_since = timezone.localtime()
            record_css_icon = ''
            events = []

        return super().get_context_data(
            record_model_name=record.__class__.__name__,
            record_css_icon=record_css_icon,
            events_since=events_since,
            event_stream=events,
            **kwargs,
        )


# ----------------------------------------------------------------------
class ActivityStreamModelViewerMixin(ActivityStreamBase):
    """
    Get activity stream for this model.
    """

    def get_event_model(self) -> type['BaseIntegerPKModel'] | type['BaseUUIDPKModel']:
        return apps.get_model(app_label=self.kwargs['app'], model_name=self.kwargs['model'])

    def get_context_data(self, **kwargs: dict[Any, Any]) -> dict[Any, Any]:
        record_model = self.get_event_model()()
        events = ConcordEventsTrail().model_events(
            event_model=record_model.pgh_event_model,
            operations=self.kwargs['operations'].split(','),
        )

        return super().get_context_data(
            record_model_name=record_model.__class__.__name__,
            record_css_icon=record_model.css_icon,
            event_stream=events,
            **kwargs,
        )



from collections import defaultdict
from datetime import datetime
import logging
from typing import Any, Literal, TypedDict, TYPE_CHECKING, Union

from django.apps import apps
from django.conf import settings
from django.db.models import Model, Q
from django.urls import reverse, NoReverseMatch

from ..models import ConcordEvents, ConcordEvent

if TYPE_CHECKING:
    from console_base.models import BaseIntegerPKModel, BaseUUIDPKModel

logger = logging.getLogger(__name__)
OPERATIONS = Literal['insert', 'update', 'delete']


class RecordEventData(TypedDict):
    model_name: str
    css_icon: str
    record: ConcordEvent
    record_url: str
    record_name: str


class OperationEventData(TypedDict):
    pk: int
    date: datetime
    model_name: str
    css_icon: str
    operation: OPERATIONS
    context: dict[str, Any]
    record_url: str
    record_name: str


# ---------------------------------------------------------------------------
class ConcordEventsTrail:
    """
    Handle retrieval and display of Record Events in an Audit Trail view.
    """

    def __init__(self, date: datetime | None = None):
        self.date = date
        self._event_models: dict[str, list[type[Model]]] = {}

    def event_models(self) -> dict:
        """
        Return all models in the App that are Record Event tracking models.
        """
        if not self._event_models:
            event_models = defaultdict(list)
            for app, models in apps.all_models.items():
                if app not in settings.AUDITABLE_APPS:
                    continue
                for name, model_class in models.items():
                    if name.endswith('event'):
                        event_models[app].append(model_class)
            self._event_models = dict(event_models)
        return self._event_models

    def unviewed_events_exist(self) -> bool:
        """
        Check to see if any events have occurred since the last date specified.
        """
        for app, models in self.event_models().items():
            for EventModel in models:
                if EventModel.objects.filter(pgh_created_at__gte=self.date).exists():
                    return True
        return False

    def unviewed_events(self) -> list[RecordEventData]:
        """
        Return all Events that occurred after the specified date for display to the User.
        """
        all_events = []

        for app, Models in self.event_models().items():
            for EventModel in Models:
                event_models = (
                    EventModel.objects.filter(
                        pgh_created_at__gte=self.date,
                    )
                    .select_related('pgh_obj')
                    .distinct('pgh_obj')
                )

                for event_model in event_models:
                    console_record = event_model.pgh_obj
                    console_record_url = console_record.get_absolute_url()

                    events = (
                        ConcordEvents.objects.tracks(console_record)
                        .filter(
                            pgh_created_at__gte=self.date,
                        )
                        .select_related(
                            'user',
                        )
                        .order_by('-pgh_created_at')
                    )

                    for event in events:
                        all_events.append(
                            RecordEventData(
                                model_name=console_record.__class__.__name__,
                                css_icon=console_record.css_icon,
                                record=event,
                                record_url=console_record_url,
                                record_name=str(console_record),
                            )
                        )

        all_events.sort(key=lambda e: e['record'].pgh_created_at, reverse=True)

        return all_events

    def record_events(
        self, target: Union['BaseIntegerPKModel', 'BaseUUIDPKModel']
    ) -> list[RecordEventData]:
        """
        Return all Events for the specified target record.
        """
        all_events = []
        url = target.get_absolute_url()

        events = (
            ConcordEvents.objects.tracks(target)
            .filter(
                pgh_created_at__gte=self.date,
            )
            .select_related(
                'user',
            )
            .order_by('-pgh_created_at')
        )

        for event in events:
            all_events.append(
                RecordEventData(
                    model_name=target.__class__.__name__,
                    css_icon=target.css_icon,
                    record=event,
                    record_url=url,
                    record_name=str(target),
                )
            )

        return all_events

    def model_events(
        self,
        event_model: ConcordEvent,
        operations: tuple[OPERATIONS] = ('delete',),
        query_params: Q | None = None,
    ) -> list[OperationEventData]:
        """
        Return Create / Delete Events for the specified model.
        """
        all_events: list[OperationEventData] = []
        if not (pgh_tracked_model := event_model.pgh_tracked_model):
            return all_events

        pgh_labels = operations or ['delete']
        event_qp: list[tuple[str, datetime | str]] = [('pgh_label__in', pgh_labels)]  # type: ignore[list-item]
        if self.date:
            event_qp.append(('pgh_created_at__gte', self.date))

        tracked_model = pgh_tracked_model()
        url_name = tracked_model.get_url_base_name
        lookup_fields = [
            'pk',
            'pgh_created_at',
            'pgh_obj_id',
            'pgh_label',
            'pgh_context',
        ]
        if 'name' in tracked_model.get_model_fields():
            lookup_fields.append('name')

        filter_params = query_params & Q(*event_qp) if query_params else Q(*event_qp)
        events = (
            event_model.objects.filter(
                filter_params,
            )
            .order_by('-pgh_created_at')
            .values_list(*lookup_fields, named=True)
        )

        for event in events:
            context: dict[str, Any] = event.pgh_context or {}
            record_name = getattr(event, 'name', '') or context.get('record_name', 'N/A')
            all_events.append(
                OperationEventData(
                    pk=event.pk,
                    date=event.pgh_created_at,
                    model_name=tracked_model.__class__.__name__,
                    operation=event.pgh_label,
                    context=context,
                    css_icon=tracked_model.css_icon,
                    record_url=get_record_url(event, url_name),
                    record_name=record_name,
                )
            )

        return all_events


def get_record_url(event: ConcordEvent, url_name: str) -> str:
    if event.pgh_label == 'delete':
        return '#'

    try:
        return reverse(url_name, kwargs={'pk': event.pgh_obj_id})  # type: ignore[attr-defined]
    except NoReverseMatch:
        return '#'



import pghistory

from django.contrib.contenttypes.models import ContentType
from django.db import models
from ..querysets import ConcordiaContentTypeManager


# ---------------------------------------------------------------------------
class RecordNameContext(models.Model):
    """
    When records are deleted, it's helpful to know
    what the record represented.

    When a model doesn't have a "name" field, add
    this mixin to save the string representation
    of the model to the context.
    """

    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        with pghistory.context(record_name=str(self)):
            return super().delete(*args, **kwargs)


# ---------------------------------------------------------------------------
class ConcordiaContentType(ContentType):
    """
    Proxy model to make it easier to add contacts via API
    RPC Methods such as changing filter password should be
    attached viewsets of this model.
    """

    IS_SYNCABLE = False

    objects = ConcordiaContentTypeManager()  # type: ignore[misc]

    class Meta:
        proxy = True
        app_label = 'concordia'

    def __str__(self) -> str:
        return self.name.title()

    @property
    def ctmodel(self) -> str:
        """For API consistency with SyncTable"""
        return str(self.model)


__all__ = (
    'RecordNameContext',
    'ConcordiaContentType',
)



import asyncio
import importlib
from signal import Signals
from types import ModuleType
from typing import Callable


def get_application(application_name: str) -> Callable | ModuleType:
    """
    Load a callable from a `module.submodule:Object.callable` string.

    From https://github.com/sivulich/mqttasgi/blob/master/mqttasgi/utils.py
    """
    module_path, object_path = application_name.split(":", 1)
    application = importlib.import_module(module_path)
    for bit in object_path.split("."):
        application = getattr(application, bit)

    if hasattr(application, "as_asgi"):
        return application.as_asgi()

    return application


def terminate(event: asyncio.Event, signal: Signals) -> None:  # pragma: no cover
    """
    Close the event loop.
    """
    event.set()
    asyncio.get_running_loop().remove_signal_handler(signal)



# mypy: disable-error-code = no-untyped-call
import sys
from typing import Any

from django.apps import apps
from django.db import models
from django.utils.module_loading import import_string

import pghistory
from pghistory import (
    ContextForeignKey,
    ContextJSONField,
    ContextUUIDField,
    DeleteEvent,
    InsertEvent,
    ObjForeignKey,
    UpdateEvent,
    config,
    constants,
    utils,
)
from pghistory.core import (
    _generate_event_model_name,
    _generate_history_field,
    _get_append_only,
    _get_context_field,
    _get_context_id_field,
    _get_obj_field,
    _validate_event_model_path,
    track,
)
import pgtrigger

from console_base.models import BaseIntegerPKModel, BaseUUIDPKModel
from .models import ConcordEvent


def cctrack(*args: Any, **kwargs) -> ConcordEvent:  # type: ignore[no-untyped-def]
    """
    Wrapper around builtin `track` decorator to always
    exclude the `created` and `modified` fields.

    If `modified` is the only changed value, (for sync purposes)
    we can skip creating a version of it.
    """
    kwargs['exclude'] = _exclude_date_fields(exclude=list(kwargs.get('exclude') or []))

    if not kwargs.get('base_model'):
        kwargs['base_model'] = ConcordEvent

    context_field = kwargs.get('context_field')
    if not context_field or context_field == pghistory.constants.UNSET:
        kwargs['context_field'] = pghistory.ContextJSONField()

    kwargs['meta'] = _set_indexes(kwargs.pop('meta', {}))
    return track(*args, **kwargs)


def cc_create_event_model(  # type: ignore[no-untyped-def]
    tracked_model: type[BaseIntegerPKModel] | type[BaseUUIDPKModel],
    *trackers,
    fields: list[str] | None = None,
    exclude: list[str] | None = None,
    obj_field: ObjForeignKey | constants.Unset = constants.UNSET,
    context_field: ContextForeignKey | ContextJSONField | constants.Unset = constants.UNSET,
    context_id_field: ContextUUIDField | constants.Unset = constants.UNSET,
    append_only: bool | constants.Unset = constants.UNSET,
    model_name: str | None = None,
    app_label: str | None = None,
    base_model: type[models.Model] | None = None,
    attrs: dict[str, Any] | None = None,
    meta: dict[str, Any] | None = None,
    abstract: bool = True,
) -> type[ConcordEvent]:
    """
    Override the builtin `create_event_model` function to always
    pass in default list of row events. Using `PGHISTORY_DEFAULT_TRACKERS`
    doesn't work because instantiated classes generate triggers polluted
    accross models. :scream:

    Create an event model.

    Instead of using [pghistory.track][], which dynamically genferates an event
    model, one can instead construct a event model themselves, which
    will also set up event tracking for the tracked model.

    Args:
        tracked_model: The model that is being tracked.
        *trackers: The event trackers. When using any tracker that inherits
            [pghistory.RowEvent][], such as [pghistory.InsertEvent][], a
            Postgres trigger will be installed that automatically stores the event
            into the generated event model. Trackers that do not inherit
            [pghistory.RowEvent][] must be manually created. If no events are
            supplied, defaults to `pghistory.InsertEvent` and `pghistory.UpdateEvent`.
        fields: The list of fields to snapshot when the event takes place. When
            no fields are provided, the entire model is snapshot when the event
            happens. Note that snapshotting of the OLD or NEW row is configured
            by the `snapshot` attribute of the `DatabaseTracker` object. Manual
            events must specify these fields during manual creation.
        exclude: Instead of providing a list of fields to snapshot, a user can
            instead provide a list of fields to not snapshot.
        obj_field: The foreign key field configuration that references the tracked object.
            Defaults to an unconstrained non-nullable foreign key. Use `None` to create a
            event model with no reference to the tracked object.
        context_field: The context field configuration. Defaults to a nullable
            unconstrained foreign key. Use `None` to avoid attaching historical context altogether.
        context_id_field: The context ID field configuration when using a ContextJSONField
            for the context_field. When using a denormalized context field, the ID
            field is used to track the UUID of the context. Use `None` to avoid using this
            field for denormalized context.
        append_only: True if the event model is protected against updates and deletes.
        model_name: Use a custom model name when the event model is generated. Otherwise
            a default name based on the tracked model and fields will be created.
        app_label: The app_label for the generated event model. Defaults to the app_label
            of the tracked model. Note, when tracking a Django model (User) or a model
            of a third-party app, one must manually specify the app_label of an internal
            app to use so that migrations work properly.
        base_model: The base model for the event model. Must inherit pghistory.models.Event.
        attrs: Additional attributes to add to the event model
        meta: Additional attributes to add to the Meta class of the event model.
        abstract: `True` if the generated model should be an abstract model.

    Returns:
        The event model class.

    Example:
        Create a custom event model:

            class MyEventModel(create_event_model(
                TrackedModel,
                pghistory.InsertEvent(),
            )):
                # Add custom indices or change default field declarations...
    """  # noqa

    # ------------------------------------------------------------------
    # Override `create_event_model` function to provide default trackers
    # that must be instantiated each time the function is called.
    # Passing in instantiated event classes via settings doesn't
    # generate correct triggers.
    if not trackers:
        trackers = [InsertEvent(), UpdateEvent(), DeleteEvent()]  # type: ignore[assignment]

    # Add a few more default variables
    exclude = _exclude_date_fields(exclude=exclude or [])
    context_field = context_field or pghistory.ContextJSONField()
    base_model = base_model or ConcordEvent
    meta = _set_indexes(meta or {})
    # ------------------------------------------------------------------

    # ------------------------------------------------------------------
    # All code below this line should be default function.
    # ------------------------------------------------------------------

    event_model = import_string("pghistory.models.Event")
    base_model = base_model or config.base_model()
    assert issubclass(base_model, event_model)

    obj_field = _get_obj_field(
        obj_field=obj_field,
        tracked_model=tracked_model,
        base_model=base_model,
        fields=fields,
    )
    context_field = _get_context_field(context_field)
    context_id_field = _get_context_id_field(context_id_field)
    append_only = _get_append_only(append_only)

    model_name = model_name or _generate_event_model_name(base_model, tracked_model, fields)
    app_label = app_label or tracked_model._meta.app_label
    _validate_event_model_path(app_label=app_label, model_name=model_name, abstract=abstract)
    app = apps.app_configs[app_label]
    app_module_name = getattr(app.module, '__name__')
    models_module = f"{app_module_name}.models"

    attrs = attrs or {}
    attrs.update({"pgh_trackers": trackers})
    meta = meta or {}
    exclude = exclude or []
    fields = fields or [f.name for f in tracked_model._meta.fields if f.name not in exclude]

    if append_only:
        meta["triggers"] = [
            *meta.get("triggers", []),
            pgtrigger.Protect(name="append_only", operation=pgtrigger.Update | pgtrigger.Delete),
        ]

    class_attrs = {
        "__module__": models_module,
        "Meta": type("Meta", (), {"abstract": abstract, "app_label": app_label, **meta}),
        "pgh_tracked_model": tracked_model,
        **{field: _generate_history_field(tracked_model, field) for field in fields},
        **attrs,
    }

    if isinstance(context_field, utils.JSONField) and context_id_field:
        class_attrs["pgh_context_id"] = context_id_field

    if context_field:
        class_attrs["pgh_context"] = context_field

    if obj_field:
        class_attrs["pgh_obj"] = obj_field

    event_model = type(model_name, (base_model,), class_attrs)
    if not abstract:
        setattr(sys.modules[models_module], model_name, event_model)

    return event_model


def _exclude_date_fields(exclude: list[str]) -> list[str]:
    date_fields = ['created', 'modified']
    if not exclude:
        return date_fields

    exclude.extend(date_fields)
    return list(set(exclude))


def _set_indexes(meta: dict) -> dict:
    indexes: list[models.Index] = meta.get('indexes') or []
    indexes.extend([
        models.Index(fields=['cid']),
        models.Index(fields=['pgh_obj']),
        models.Index(fields=['pgh_created_at']),
    ])
    meta['indexes'] = indexes

    return meta


__all__ = (
    'cctrack',
    'cc_create_event_model',
)



# Tasks for Publishing messages to Concordia Message Queue
from celery import shared_task
from celery_singleton.exceptions import DuplicateTaskError
import logging

from django.conf import settings

from ..acls import skip_publication
from ..channels import ConcordTriggerChannel, ROOM_CHANNEL_REGISTRY
from ..choices import CONTEXT, CRUD, ORIGIN_SERVER
from ..constants import OPERATION_KEY, PORTAL_RECORD_PK_KEY
from ..events.operations import mark_events_synced, mq_payload_from_events
from ..mq.helpers import publish_concord_message

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
@shared_task
def sync_subscriber_publish_local_crud_event_to_concord_room(room_name: str, data: dict) -> None:
    """
    When ConcordEvents are made on Sync Subscribers, post message to
    the Sync Publisher Channel. The Sync Publisher will be responsible to
    ensure that the change is published to all relevant Subscribers.
    """
    logger.info(
        f'sync_subscriber_publish_local_crud_event_to_concord_room {room_name}. Data: {data}'
    )
    return
    channel = ROOM_CHANNEL_REGISTRY[room_name]
    record = channel.model.pgh_tracked_model.objects.filter(id=data[PORTAL_RECORD_PK_KEY]).first()

    # Recheck for publication status, now that we have a record object to inspect
    if skip_publication(record=record):
        return

    payload, record_events = mq_payload_from_events(record=record, serializer=channel.serializer)
    payload[CONTEXT.Origin] = data[CONTEXT.Event]

    # Include the Event Context ID(s) so that the Publisher can return then
    # in the Response back, and then we'll mark the ConcordEvents as synced.
    payload[CONTEXT.Origin][CONTEXT.Keys] = [re.pgh_context_id for re in record_events]

    rooms = channel.publish_room_names(record)
    result = publish_concord_message(rooms=rooms, payload=payload)

    published_status = [m.is_published() for m in result]
    logger.info(
        f'Concordia MQ Subscriber->Willow Event Room {rooms!r}. '
        f'Data: {payload!r}. Published: {published_status}'
    )


# ---------------------------------------------------------------------------
@shared_task
def sync_publisher_publish_local_crud_event_to_concord_room(room_name: str, data: dict) -> None:
    """
    When ConcordEvents are made on the Willow Sync Publisher,
    post message to all the Subscribers that need this record.
    """
    logger.info(
        f'sync_publisher_publish_local_crud_event_to_concord_room {room_name}. Data: {data}'
    )
    return
    channel = ROOM_CHANNEL_REGISTRY[room_name]
    record = channel.model.pgh_tracked_model.objects.filter(id=data[PORTAL_RECORD_PK_KEY]).first()

    # Recheck for publication status, now that we have a record object to inspect
    if skip_publication(record=record):
        return

    payload, record_events = mq_payload_from_events(record=record, serializer=channel.serializer)
    payload[CONTEXT.Origin] = data[CONTEXT.Event]
    context_ids = [re.pgh_context_id for re in record_events]

    try:
        # Appliance CID gets added in middleware, so it should always be present,
        # but don't crash if it isn't.
        origin_concord_mq = data[CONTEXT.Event][ORIGIN_SERVER.CID]
    except KeyError:
        logger.error(f'Unable to get origin Concord MQ CID from {payload}')
        origin_concord_mq = ''

    if subscriber_rooms := channel.publish_room_names(record, origin_concord_mq):
        result = publish_concord_message(rooms=subscriber_rooms, payload=payload)
        mark_synced = bool(result)
        msg = f'Published {subscriber_rooms!r}. Payload: {payload!r}. Result: {result}'
    else:
        mark_synced = True
        msg = f'No subscribers required payload: {payload} from {room_name}'

    logger.info(f'Concordia MQ Willow->Subscriber - {msg}')

    # Mark the events as synced to remove them from future sync operations.
    if mark_synced:
        mark_events_synced(channel.model.objects.filter(pgh_context_id__in=context_ids))


# ---------------------------------------------------------------------------
@shared_task
def publish_local_delete_event_to_concord_room(data: dict) -> None:
    """
    When local Concord Events deletion records are made, post message
    to Publisher Channel. A record that can be deleted on one Subscriber
    could still be relevant to other subscribers, so the Publisher
    will decide whether to send a message to the Subscribers.
    """
    logger.info(f'publish_local_delete_event_to_concord_room. Data: {data}')
    return
    # channel = ROOM_CHANNEL_REGISTRY[room_name]
    # payload = {
    #     'name': data.get('name', 'N/A'),
    #     'cid': data['cid'],
    #     CONTEXT.Origin: data[CONTEXT.Event],
    # }
    #
    # # Record the Canonical ID so it's never used again, unless specifically cleared.
    # if settings.SYNC_PUBLISHER:
    #     CanonicalIdDeleteLog.objects.record_deletion(record=channel.model, data=payload)
    #
    # room = channel.get_room_name()
    # result = publish_concord_message(rooms=[room], payload=payload)
    #
    # logger.info(
    #     f'Concordia MQ to Publisher Delete Room {room!r}. Data: {payload!r}. Result: {result}'
    # )


# ---------------------------------------------------------------------------
def schedule_publish_local_events_to_concord_room(
    channel: type[ConcordTriggerChannel],
    data: dict,
) -> None:
    """
    Wrapper to choose correct Publish task,
    handling DuplicateTaskError exception.
    """
    logger.info(f'schedule_publish_local_events_to_concord_room {channel}. Data: {data}')
    return
    if skip_publication(data):
        return

    # For task API consistency, ensure that data context is always dict,
    # since data payload may not always be generated from ConcordEvent.
    if (_ := data.get(CONTEXT.Event)) is None:
        data[CONTEXT.Event] = {}

    if data[CONTEXT.Event].get(OPERATION_KEY) == CRUD.Delete:
        task_function = publish_local_delete_event_to_concord_room
    else:
        if settings.SYNC_PUBLISHER:
            task_function = sync_publisher_publish_local_crud_event_to_concord_room
        else:
            task_function = sync_subscriber_publish_local_crud_event_to_concord_room

    try:
        result = task_function.apply_async(
            expires=settings.CONCORD_MQ_ROOM_POST_TASK_EXPIRATION,
            kwargs={
                'data': data,
            },
        )
        return True, result
    except DuplicateTaskError:
        msg = f'Sync task already scheduled for Channel: {channel}'
        logger.error(msg)
        return False, msg


__all__ = (
    'sync_subscriber_publish_local_crud_event_to_concord_room',
    'publish_local_delete_event_to_concord_room',
    'schedule_publish_local_events_to_concord_room',
)



import logging
from typing import Optional, TYPE_CHECKING
from django.conf import settings
from .choices import CONTEXT, SYNCMODES

if TYPE_CHECKING:
    from .models import SyncableModel

from .constants import (
    SYNCMODE_KEY,
    SYNC_COMPLETED_KEY,
)

logger = logging.getLogger(__name__)


def skip_publication(
    record_data: dict | None = None,
    record: Optional['SyncableModel'] = None,
) -> bool:
    """
    Some events should not be published back to the Concordia MQ.
    Examine the event data to determine if the change should be
    published to the MQ or not.
    """
    if not record_data:
        return False

    try:
        if record_data[SYNCMODE_KEY] == SYNCMODES.Local:
            return False
    except KeyError:
        pass

    try:
        if record_data[CONTEXT.Origin][SYNC_COMPLETED_KEY]:
            return True
    except KeyError:
        pass

    if not settings.SYNC_ENABLED:
        logger.debug('Skip publishing %s to "%s" - sync is disabled', record)
        return True

    if not record:
        return False

    if not record.is_syncable():
        return True

    return False



# Tasks for syncing via REST API


# @shared_task(base=LCSingleton, lock_expiry=43200, raise_on_duplicate=True)
# def synchronize_all(batches, **kwargs):
#    return 'done'


# __all__ = [
#    'synchronize_all',
# ]



import logging

logger = logging.getLogger(__name__)
CONTENT_TYPE_COMPRESSED_JSON = 'compressed/json'
CONTENT_TYPE_JSON = 'json'


class BaseConcordiaClient:
    """
    Create base Clarion client for typical portal operations.
    """

    def publish(
        self,
        room: str,
        payload: dict | None = None,
    ) -> dict:
        print(f'Publish {payload} to {room}')
        return payload or {}


class ConcordiaClient(BaseConcordiaClient):
    """
    General client class to use for Publishing / Subscribing to Concord MQ rooms.
    """

    ...


__all__ = (
    'BaseConcordiaClient',
    'ConcordiaClient',
)



#!/usr/bin/env python3
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from concordia import __version__ as VERSION


def read_me():
    with open('README.md') as read:
        return read.read()


def requirements():
    with open('requirements.txt') as req:
        print([r.strip() for r in req.readlines()])
        return [r.strip() for r in req.readlines()]


setup(
    name='concordia',
    version=VERSION,
    description="Concordia Sync functionality for Compass Consoles written in Django.",
    long_description=read_me(),
    author='Dave Burkholder',
    author_email='dave@compassfoundation.io',
    url='https://code.compassfoundation.io/console-modules/concordia',
    packages=['concordia'],
    include_package_data=True,
    install_requires=requirements(),
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'Operating System :: POSIX :: Linux',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
    ],
)



from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes.fields import GenericForeignKey
from django.db import models
from django.db.models import (
    CheckConstraint,
    Index,
    Q,
    UUIDField,
)
from django.utils.functional import cached_property
from console_base.models import (  # type: ignore[attr-defined]
    BaseUUIDPKModel,
    CanonicalIdField,
    LCUniqueConstraint,
)
from ..querysets import (
    CanonicalIdDeleteLogManager,
    CanonicalIdMapManager,
    CanonicalIdSwapManager,
    SyncQuerySet,
    SYNC_TABLES_Q,
)


# ---------------------------------------------------------------------------
class GenericCanonicalIdTrackerMixin(BaseUUIDPKModel):
    """
    Mixin class to track Canonical ID record changes with
    Generic FK reference to database record.
    """

    cid = CanonicalIdField()
    name = models.TextField()
    table = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        limit_choices_to=SYNC_TABLES_Q,
    )
    object_id = models.TextField(null=True, blank=True)
    record_object = GenericForeignKey(ct_field='table')

    class Meta:
        abstract = True
        indexes = [
            Index(fields=['object_id'], name='generic_record_pk_idx'),
            Index(fields=['cid'], name='cid_tracker_cid_idx'),
        ]

    @cached_property
    def app_label(self) -> str:
        return str(self.table.app_label)

    @cached_property
    def table_name(self) -> str:
        return str(self.table.model)


# ---------------------------------------------------------------------------
class CanonicalIdSwap(GenericCanonicalIdTrackerMixin):
    """
    When a Canonical ID needs to be changed on a record, keep
    a log of the prior value so that we can be sure to enforce
    that sync operations don't restore the old value.
    """

    css_icon = ''
    md_description = 'Swap Old for New Canonical ID'

    old = UUIDField()

    objects = CanonicalIdSwapManager()

    class Meta:
        app_label = 'concordia'
        verbose_name_plural = 'canonical id map'

        constraints = [
            CheckConstraint(
                name='old_cid_differs_from_new_cid',
                condition=Q(cid=models.F('old'), _negated=True),
                violation_error_message='Old CID must differ from new CID',
            ),
            LCUniqueConstraint(
                fields=('object_id', 'old'),
                name='old_canonical_id_unique',
                violation_error_message='Record swapping this old Canonical ID already exists.',
            ),
        ]

    def apply(self):
        """
        Apply this Canonical ID swap to the local database record.
        """
        try:
            record: BaseUUIDPKModel | None = self.record_object
            if record and record.cid != self.cid:
                record.cid = self.cid
                record.save(update_fields=['cid'])
        except models.Model.DoesNotExist:  # type: ignore[attr-defined]
            pass


# ---------------------------------------------------------------------------
class CanonicalIdMap(GenericCanonicalIdTrackerMixin):
    """
    Records have a Canonical ID field that should identify the record
    throughout all portal consoles. However, if it's not possible to
    have a record's CID be globally unique, this model will store the
    relationship of the local CID to the global CID.

    Examples might be records synced with an Active Directory server,
    where the AD server's UUID value is used as CID. It's not possible
    to change the AD server's value, so a mapping entry can be made.
    """

    css_icon = ''
    md_description = 'Map Local Canonical ID to the Global Canonical ID'

    local = UUIDField()

    objects = CanonicalIdMapManager()

    class Meta:
        app_label = 'concordia'
        verbose_name_plural = 'canonical id map'

        constraints = [
            CheckConstraint(
                name='local_cid_differs_from_global_cid',
                condition=Q(cid=models.F('local'), _negated=True),
                violation_error_message='Local CID must differ from global CID',
            ),
            LCUniqueConstraint(
                fields=('object_id', 'local'),
                name='local_canonical_id_unique',
                condition=Q(object_id__isnull=False),
                violation_error_message='Local record mapping to Canonical ID already exists.',
            ),
        ]

    def apply(self):
        """
        Apply this Canonical ID Map to the local database record.
        """
        try:
            record: BaseUUIDPKModel | None = self.record_object
            if record and record.cid != self.local:
                self.local = record.cid
                record.save(update_fields=['local'])
        except models.Model.DoesNotExist:  # type: ignore[attr-defined]
            pass


# ---------------------------------------------------------------------------
class CanonicalIdDeleteLog(GenericCanonicalIdTrackerMixin):
    """
    Canonical IDs that have been deleted cannot be re-used.

    These records will normally be created on Sync Publishers
    to ensure that Sync Subscribers or other clients cannot
    re-create a deleted record automatically. If a record
    is to be recreated, a new Canonical ID can be used,
    or the CanonicalIdDeleteLog record be deleted.
    """

    css_icon = ''
    md_description = 'Canonical ID Deletion Record'

    objects = CanonicalIdDeleteLogManager()

    class Meta:
        app_label = 'concordia'
        verbose_name_plural = 'canonical id delete'

        constraints = [
            LCUniqueConstraint(
                fields=['cid'],
                name='deleted_canonical_id_unique',
                violation_error_message='Delete record for this Canonical ID already exists.',
            ),
        ]

    def apply(self):
        """
        Ensure that this record is deleted.
        """
        if not self.record_object:
            return

        self.record_object.delete()


# ---------------------------------------------------------------------------
class SyncableModel(BaseUUIDPKModel):
    """
    Mixin class to add sync methods for type hints
    """

    LOCAL_CONSTRAINT_FIELDS: frozenset = frozenset()
    cid = CanonicalIdField()

    objects = SyncQuerySet.as_manager()

    class Meta:
        abstract = True

    def is_syncable(self) -> bool:
        return True


__all__ = (
    'GenericCanonicalIdTrackerMixin',
    'CanonicalIdSwap',
    'CanonicalIdMap',
    'CanonicalIdDeleteLog',
    'SyncableModel',
)



from console_base.api.routers import LCSimpleRouter
from . import views as cv

router = LCSimpleRouter()
router.register(
    r'concordia/canonicaliddeletelog',
    cv.CanonicalIdDeleteLogViewSet,
    basename='api-canonicaliddeletelog'
)
router.register(
    r'concordia/canonicalidmap', cv.CanonicalIdMapViewSet, basename='api-canonicalidmap'
)
router.register(
    r'concordia/canonicalidswap', cv.CanonicalIdSwapViewSet, basename='api-canonicalidswap'
)



import logging
import pgtrigger
from typing import Any, Iterable, TYPE_CHECKING

from console_base.api.serializers import LCModelSerializer
from django.conf import settings
from django.http import HttpRequest
from rest_framework.request import Request

from ..choices import SYNCMODES
from ..constants import SYNC_COMPLETED_KEY, SYNCMODE_KEY
from ..models import ConcordEvents
from ..typehints import EventData, Status

if TYPE_CHECKING:
    from ..models import SyncableModel
    from concordia_tests.latchstring.models import User

logger = logging.getLogger(__name__)

http_request = HttpRequest()
http_request.META = {
    'SERVER_NAME': settings.HOSTNAME,
    'SERVER_PORT': settings.EXTERNAL_PORT,
}


# ---------------------------------------------------------------------------
def mq_payload_from_events(
    record: 'SyncableModel',
    serializer: type[LCModelSerializer],
) -> EventData:
    """
    Get all unsynced ConsoleEvent records for the
    specified Console record that was just changed.

    Merge event data into Concordia MQ payload that
    will be published to the relevant MQ room.
    """
    serialized = serialized_record_data(record, serializer)
    events = ConcordEvents.objects.tracks(record).unsynced().order_by('-pgh_created_at')

    if not record_is_syncable(record):
        return EventData({}, events)

    event_data = []
    for event in events:
        record_diff: dict[str, Any] = event.pgh_diff or {}
        if record_diff:
            diff = {'modified': event.pgh_created_at}
            # Extract new value from record diff and ignore the old value
            for field, values in record_diff.items():
                diff[field] = values[1]

            for field in constraint_requirements(record_diff, record):
                try:
                    diff[field] = serialized[field]
                except KeyError:
                    logger.error(f'Unexpected error! Field: {field} not found in serialized data')
                    continue

        else:
            diff = serialized
            diff['modified'] = event.pgh_created_at

        event_data.append(diff)

    merged_data = merge_event_diff_data(event_data)

    # Ensure that `cid` field is present so the
    # record can be retrieved on Sync Subscriber
    if 'cid' not in merged_data:
        merged_data['cid'] = {'value': record.cid}

    return EventData(convert_pks_to_cids(record, serialized, merged_data), events)


# ---------------------------------------------------------------------------
def convert_pks_to_cids(record: 'SyncableModel', serialized: dict, merged_data: dict) -> dict:
    """
    Convert Foreign Key field value to Canonical ID for REST API compatibility.
    """
    related_fields = [f.name for f in record._meta.fields if f.related_model]

    for field in related_fields:
        if field in merged_data:
            merged_data[field]['value'] = serialized[field]

    return merged_data


# ---------------------------------------------------------------------------
def serialized_record_data(record: 'SyncableModel', serializer: type[LCModelSerializer]) -> dict:
    """
    Serialize record to ensure that data is ready for API use,
    such as FK fields getting converted to Canonical IDs.
    """
    api_request = Request(request=http_request)
    rs = serializer(instance=record, context={'request': api_request})
    readonly_fields = rs.get_readonly_fields()
    return {k: v for k, v in rs.data.items() if k not in readonly_fields}


# ---------------------------------------------------------------------------
def constraint_requirements(record_diff: dict, record: 'SyncableModel') -> set[str]:
    """
    Multi-field constraints require that all fields be present to perform validation.
    So when a ConcordEvent diff doesn't include all the necessary fields, they must be
    added from serialized data and included in the MQ payload, if the values didn't change.

    :param record_diff: The fields that were changed in the local editing event.
    :param record: The database record in question.
    :return: list tuples of fields from Constraints that match against multiple fields.
        Example return values:
          * [('name', 'patterntype', 'category')]
          * [('category', 'action_group')]
          * [('policy', 'code', 'attribute'), ('code', 'attribute')]
    """
    extra_fields: set[str] = set()
    if not (record_constraints := record.get_constraints()):
        return extra_fields

    diff_fields = set(record_diff.keys())

    for model_class, model_constraints in record_constraints:
        # model_constraints are lists of tuples
        # [('name', 'patterntype', 'category')]
        # [('category', 'action_group')]
        # [('policy', 'code', 'attribute'), ('code', 'attribute')]
        for constraint in model_constraints:
            try:
                if len(constraint.fields) > 1:  # type: ignore
                    if not diff_fields.issuperset(set(constraint.fields)):  # type: ignore[attr-defined]
                        extra_fields.update(constraint.fields)  # type: ignore[attr-defined]
            except AttributeError:
                continue

    try:
        return extra_fields - record.LOCAL_CONSTRAINT_FIELDS
    except AttributeError:
        return extra_fields


# ---------------------------------------------------------------------------
def mark_events_synced(events: Iterable[ConcordEvents]) -> None:
    """
    Mark all the Concord Events as Synced with Publisher.
    """
    with pgtrigger.ignore():
        for event in events:
            context: dict[str, Any] = event.pgh_context or {}
            context[SYNC_COMPLETED_KEY] = True
            event.pgh_context = context
            event.save(update_fields=['pgh_context'])


# ---------------------------------------------------------------------------
def record_is_syncable(record: 'SyncableModel') -> bool:
    """
    Helper function to determine if a record should be synced or not.
    """
    if not record:
        return False

    try:
        return record.is_syncable()
    except AttributeError:
        pass

    return getattr(record, SYNCMODE_KEY, SYNCMODES.Sync) == SYNCMODES.Sync


# ---------------------------------------------------------------------------
def merge_event_diff_data(events: list[dict]) -> dict:
    """
    Merge all the fields from the list of diff dictionaries,
    into a dict, keeping the latest value.

    This data dict will then be posted to a Sync Publisher or Subscriber
    where each field can be compared, to ensure that the sync operation
    does not stomp on newer changes.

    data = [
        {'name': 'Allen Oxford Exc',
        'comments': 'They are in Ohio',
        'modified': datetime.datetime(2023, 1, 28, 19, 33, 46, 156944, tzinfo=timezone.utc)},

        {'name': 'Allen Oxford Tractors',
        'modified': datetime.datetime(2023, 1, 28, 19, 35, 59, 226902, tzinfo=timezone.utc)}
    ]

    :returns
    {
        'name': {
            'value': 'Allen Oxford Tractors',
            'modified': datetime.datetime(2023, 1, 28, 19, 35, 59, 226902, tzinfo=timezone.utc),
        },
        'comments': {
            'value': 'They are in Ohio',
            'modified': datetime.datetime(2023, 1, 28, 19, 33, 46, 156944, tzinfo=timezone.utc),
        },
    }
    """
    if len(events) > 1:
        # Ensure sort order from newest to oldest, to skip older fields
        events.sort(key=lambda evt: evt['modified'], reverse=True)

    merged_event_data = {}
    for event_diff in events:
        modified = event_diff['modified']
        for field, value in event_diff.items():
            if field not in merged_event_data:
                merged_event_data[field] = {'value': value, 'modified': modified}

    merged_event_data.pop('modified', None)

    return merged_event_data


# ---------------------------------------------------------------------------
def prune_record_fields(record: 'SyncableModel', mq_payload_data: dict) -> dict:
    """
    Takes a Message Queue payload dict of fields with `value` / `modified`
    values, and compares against ConcordEvent history. Return all fields
    in the payload that are newer than the last change for that field.

    data = {
        'name': {
            'value': 'Allen Oxford Tractors',
            'modified': datetime.datetime(2023, 1, 28, 19, 35, 59, 226902, tzinfo=timezone.utc),
        },
        'comments': {
            'value': 'They are in Ohio',
            'modified': datetime.datetime(2023, 1, 28, 19, 33, 46, 156944, tzinfo=timezone.utc),
        },
    }
    """
    record_events = ConcordEvents.objects.tracks(record)
    pruned_fields = {}

    for field, params in mq_payload_data.items():
        # payload will contain CID field value for record retrieval,
        # in which case it won't have the "modified" field, so skip.
        if not (modified := params.get('modified')):
            continue
        if not record_events.filter(
            pgh_created_at__gte=modified,
            pgh_diff__has_key=field,
        ).exists():
            pruned_fields[field] = params['value']

    return pruned_fields


# ---------------------------------------------------------------------------
def persist_changed_fields(
    record: 'SyncableModel',
    serializer: type[LCModelSerializer],
    cleaned_data: dict,
    user: 'User',
) -> Status:
    """
    Save changed fields to database. All relevancy
    checks should be performed before this point.
    """
    if record:
        cleaned_data.pop('cid', None)

    if not cleaned_data:
        return Status(True, {'message': 'No changes to save'})

    api_request = Request(request=http_request)
    api_request.user = user

    model_serializer = serializer()
    readonly_fields = set(model_serializer.get_readonly_fields())
    writable_fields = set(model_serializer.get_fields().keys()) - readonly_fields
    partial_update = bool(writable_fields.difference(set(cleaned_data.keys())))

    # TODO extend serializer to generate MQ-oriented errors
    rs = serializer(
        data=cleaned_data,
        instance=record,
        context={'request': api_request},
        partial=partial_update,
    )

    if not rs.is_valid():
        logger.info(f'Concordia invalid serializer {rs.errors}')
        return Status(False, {'errors': rs.errors})

    updated_record = rs.save()

    return Status(True, {'message': f'Sync operation success for: {updated_record}'})


__all__ = (
    'constraint_requirements',
    'mq_payload_from_events',
    'merge_event_diff_data',
    'mark_events_synced',
    'convert_pks_to_cids',
    'serialized_record_data',
    'prune_record_fields',
    'persist_changed_fields',
)



from typing import Any, Optional, TYPE_CHECKING
from django.conf import settings
from django.contrib.auth import get_user_model

from ..models import SyncableModel
from .client import ConcordiaClient
from ..typehints import Record

if TYPE_CHECKING:
    from concordia_tests.latchstring.models import User
else:
    User = get_user_model()


# ---------------------------------------------------------------------------
def publish_concord_message(rooms: list[str], payload: dict) -> list[dict[Any, Any]]:
    """
    Helper method to publish a message to Concord MQ.
    """
    client = ConcordiaClient()
    results = []

    for room in rooms:
        result = client.publish(room=room, payload=payload)
        results.append(result)

    return results


# ---------------------------------------------------------------------------
def load_record(model: SyncableModel | None, cid: str, context: dict) -> Record:
    """
    Load model record from Concord MQ message data.

    If this is a Sync Publisher, load the User from the context,
    to ensure that the Subscriber User that published the message
    has access to the record.
    """
    # The Concordia client should fail registration if the user
    # is not found, but as a sanity check, ensure no perms are
    # granted if the User is not found.
    if not model or not (user := load_user_from_context(context)):
        return Record(None, None, False)  # type: ignore

    RecordModel = model

    # TODO consider some way of tracking the Publisher's user CID on Subscribers
    if not settings.SYNC_PUBLISHER:
        return Record(user, RecordModel.objects.filter(cid=cid).first(), True)

    # If the record exists and the user queried it successfully,
    # then the Subscriber user is permitted to publish changes.
    if record := RecordModel.objects.tenant(user).filter(cid=cid).first():  # type: ignore[attr-defined]
        return Record(user, record, True)

    # If untenanted query confirms the record doesn't exist, `create` operation is permissible.
    return Record(user, None, permitted=not RecordModel.objects.filter(cid=cid).exists())


# ---------------------------------------------------------------------------
def load_user_from_context(context: dict) -> Optional['User']:
    """
    Expects a context dict from a pghistory event.
    Return user object from context dict.
    """
    try:
        return User.objects.get(cid=context['user_cid'])
    except (KeyError, User.DoesNotExist):
        pass


__all__ = (
    'publish_concord_message',
    'load_record',
    'load_user_from_context',
)



from django_filters import rest_framework as filters

from console_base.filters import NameFilter

from concordia.models import CanonicalIdDeleteLog, CanonicalIdMap, CanonicalIdSwap


# ---------------------------------------------------------------------------
class CanonicalIdDeleteLogFilter(NameFilter):
    table = filters.CharFilter(field_name='table', lookup_expr='exact')
    app_label = filters.CharFilter(
        field_name='table__app_label',
        lookup_expr='exact',
        label='App Name',
    )
    table_name = filters.CharFilter(
        field_name='table__model',
        lookup_expr='exact',
        label='Table Name',
    )

    class Meta(NameFilter.Meta):
        model = CanonicalIdDeleteLog
        fields = NameFilter.Meta.fields + (
            'table',
            'app_label',
            'table_name',
        )


# ---------------------------------------------------------------------------
class CanonicalIdMapFilter(CanonicalIdDeleteLogFilter):
    class Meta(CanonicalIdDeleteLogFilter.Meta):
        model = CanonicalIdMap


# ---------------------------------------------------------------------------
class CanonicalIdSwapFilter(CanonicalIdDeleteLogFilter):
    class Meta(CanonicalIdDeleteLogFilter.Meta):
        model = CanonicalIdSwap



from dataclasses import dataclass
from lchttp.json import json_loads
from pgpubsub.channel import BaseChannel
from pgpubsub.listeners import _trigger_action_listener
import pgtrigger
from typing import Callable, Optional, TYPE_CHECKING
from uuid import UUID

from django.conf import settings
from django.utils.module_loading import import_string

from .rooms import RoomChannelRegistry

if TYPE_CHECKING:
    from .models import ConcordEvent, SyncableModel
    from console_base.api.serializers import LCModelSerializer

ROOM_CHANNEL_REGISTRY = RoomChannelRegistry()
try:
    tenanted_publish_rooms = import_string(settings.CONCORD_MQ_TENANTED_PUBLISH_ROOMS)
except (AttributeError, ImportError):
    # print('settings does not define CONCORD_MQ_TENANTED_PUBLISH_ROOMS')

    def tenanted_publish_rooms(*args, **kwargs):
        return []


@dataclass
class ConcordTriggerChannel(BaseChannel):
    """
    Trigger for Concord Event history, deserializing
    JSON record data to a dict.
    """

    model: 'ConcordEvent'
    new: dict

    # attributes for handling Clarion messages
    serializer: type['LCModelSerializer']

    @classmethod
    def room_name(cls) -> str:
        if (name := cls.model._meta.model_name or '').startswith('concord_'):
            return name[8:]
        return name

    @classmethod
    def listen_safe_name(cls) -> str:
        """
        Postgres LISTEN protocol accepts channel names
        which are at most 63 characters long.
        """
        model_name = cls.model().__class__.__name__.lower()  # type: ignore[operator]
        if model_name.endswith('concordevent'):
            model_name = model_name[:-12]
        return f'concord_{model_name}'

    @classmethod
    def deserialize(cls, payload: dict | str) -> dict:
        """
        Override method to use faster JSON parsing and to return
        a dict that will correctly initialize Channel dataclass.
        """
        if isinstance(payload, str):
            payload = json_loads(payload)

        return {
            'new': payload['new'],  # type: ignore
            'model': cls.model,
            'serializer': cls.serializer,
        }

    @classmethod
    def register(cls, callback: Callable) -> None:
        super().register(callback)

        # TODO design way to handle registry uniquely
        # if cls.room_name in ROOM_CHANNEL_REGISTRY:
        #     raise ValueError(f'{cls.room_name} is already found in the registry')
        #
        # ROOM_CHANNEL_REGISTRY[cls.room_name] = cls

    @classmethod
    def publish_room_names(
        cls,
        record: Optional['SyncableModel'] = None,
        origin_appliance_cid: str | UUID = '',
    ) -> list[str]:
        """
        Return all Room Names to Publish new Messages.

        When we're the Sync Publisher, publish to all Sync
        Subscribers that contain the relevant tenant.

        When we're the Sync Subscriber, only publish message
        to the Sync Publisher, which will handle distribution.

        When a message originated from a Subscriber appliance, we
        don't want to include that appliance's room in the list.
        """
        if settings.SYNC_PUBLISHER:
            return tenanted_publish_rooms(record, cls.room_name, origin_appliance_cid)

        return [f'{settings.PUBLISHER_ROOM_PREFIX}/{cls.room_name}/{settings.CONCORD_MQ_TENANT}']

    @classmethod
    def subscribe_room_name(cls) -> str:
        """
        Return complete room name to use when Subscribing to this Room.
        """
        # Sync Publishers should only subscribe to Rooms specifically prefixed for them
        if settings.SYNC_PUBLISHER:
            raise NotImplementedError('sync publisher room name not calculated yet')

        raise NotImplementedError('subscribe room name not calculated yet')

    @classmethod
    def deploy_room_name(cls) -> str:
        """
        Return room name prefix for Sync Publishers to use
        when broadcasting / publishing successful changes on
        the Publisher to all Sync Subscribers.
        """
        raise NotImplementedError('deploy room name not calculated yet')


def post_update_event_listener(channel: type[ConcordTriggerChannel] | str):  # type: ignore[no-untyped-def]
    return _trigger_action_listener(
        channel=channel,
        when=pgtrigger.After,
        operation=pgtrigger.Update,
    )


def post_insert_event_listener(channel: type[ConcordTriggerChannel] | str):  # type: ignore[no-untyped-def]
    return _trigger_action_listener(
        channel=channel,
        when=pgtrigger.After,
        operation=pgtrigger.Insert,
    )


def post_delete_event_listener(channel: type[ConcordTriggerChannel] | str):  # type: ignore[no-untyped-def]
    return _trigger_action_listener(
        channel=channel,
        when=pgtrigger.After,
        operation=pgtrigger.Delete,
    )


__all__ = (
    'ConcordTriggerChannel',
    'post_update_event_listener',
    'post_insert_event_listener',
    'post_delete_event_listener',
    'ROOM_CHANNEL_REGISTRY',
    'RoomChannelRegistry',
)



# Tasks for responding to messages from Concordia Message Queue Subscriptions
from celery import shared_task
import logging
import pghistory
from typing import Any

from django.conf import settings
from django.db import DatabaseError

from ..channels import (
    ConcordTriggerChannel,
    ROOM_CHANNEL_REGISTRY,
)
from ..choices import CONTEXT
from ..constants import (
    SYNC_COMPLETED_KEY,
)
from ..events.operations import (
    mark_events_synced,
    prune_record_fields,
    persist_changed_fields,
)
from ..mq.helpers import load_record
from ..typehints import Status

logger = logging.getLogger(__name__)


# ---------------------------------------------------------------------------
@shared_task
def persist_concord_room_message_to_database(
    message_room: str,
    payload: dict,
) -> Status:
    """
    Take a ConcordiaMQ message and save it to local database.
    """
    channel = ROOM_CHANNEL_REGISTRY[message_room]

    # Include context from upstream server and attach it to this instance
    # to retain clearer picture of where the record change originated.
    context: dict[str, Any] = payload.pop(CONTEXT.Origin) or {}

    # If the Sync Publisher is receiving a message from a Subscriber, sync
    # can't be marked "completed" until messages are broadcast to Subscribers
    context[SYNC_COMPLETED_KEY] = not settings.SYNC_PUBLISHER

    room = channel.room_name()
    serializer = channel.serializer
    Model = channel.model.pgh_tracked_model

    cid = payload['cid']['value']
    record = load_record(Model, cid, context)
    if not record.permitted:
        msg = f'User not permitted to change record with CID: {cid!r} on Room: {room!r}'
        logger.error(msg)
        return Status(False, {'message': msg})

    if not record.obj:
        return Status(False, {'message': 'No record object found'})

    # Lookup any unsynced records so the incoming data can be merged with local unsynced data
    # event_data, record_events = mq_payload_from_events(record=record.obj, serializer=serializer)

    logger.info(f'Concordia MQ msg - {room!r} data: {payload!r} context: {context}')

    # Record actor was queried by CID; assign PK for
    # local FK relations to work on ConcordEvents model
    if record.editor:
        context['user'] = record.editor.pk

    with pghistory.context(**context):
        status = persist_changed_fields(
            record=record.obj,
            serializer=serializer,
            cleaned_data=prune_record_fields(record.obj, payload),
            user=record.editor,
        )

    return status


# ---------------------------------------------------------------------------
@shared_task
def persist_concord_room_deletion_message_to_database(
    message_room: str,
    payload: dict,
) -> Status:
    """
    Take a ConcordiaMQ deletion message and ensure it's deleted.
    """
    channel = ROOM_CHANNEL_REGISTRY[message_room]

    # Include context from upstream server and attach it to this instance
    # to retain clearer picture of where the record change originated.
    context = payload[CONTEXT.Origin]

    # If the Sync Publisher is receiving a message from a Subscriber, sync
    # can't be marked "completed" until messages are broadcast to Subscribers
    context[SYNC_COMPLETED_KEY] = not settings.SYNC_PUBLISHER

    cid = payload['cid']['value']
    record = load_record(channel.model.pgh_tracked_model, cid, context)
    if not record.obj:
        return Status(True, {'message': f'No record found for {cid}'})

    # Record actor was queried by CID; assign PK for
    # local FK relations to work on ConcordEvents model
    if record.editor:
        context['user'] = record.editor.pk

    with pghistory.context(**context):
        try:
            # Don't force deletion if this record is a Sync Publisher
            # since other records may depend on its existence.
            record.obj.delete(force=not settings.SYNC_PUBLISHER)
            status = Status(True, {'message': f'Record {record.obj} was deleted'})
        except DatabaseError as e:
            logger.error('Unable to delete %s - %s', (record.obj, e))
            status = Status(False, {'message': str(e)})

    return status


# ---------------------------------------------------------------------------
def confirm_event_synced(channel: type[ConcordTriggerChannel], data: dict) -> Status:
    """
    Confirm that the ConcordEvent record has been synchronized.
    """
    EventModel = channel.model
    events = EventModel.objects.unsynced().filter(pgh_context_id__in=data[CONTEXT.Keys])  # type: ignore[attr-defined]

    try:
        mark_events_synced(events)
        msg = f'Marked {len(events)} as synced in Topic: {channel.room_name()}'
        return Status(True, {'message': msg})
    except Exception as e:
        logger.exception('Unable to confirm synced events %s: %s', (channel, e))
        return Status(False, {'message': str(e)})


__all__ = (
    'persist_concord_room_message_to_database',
    'persist_concord_room_deletion_message_to_database',
    'confirm_event_synced',
)



from typing import TYPE_CHECKING

from django.conf import settings
from django.db.models import Model

if TYPE_CHECKING:
    from .channels import ConcordTriggerChannel


class RoomChannelRegistry(dict):
    """
    Map the base names of Concord MQ Rooms to PubSub Channel names.

    Concord MQ room names will have various forms for these use cases:
       * Console Subscribers POSTing local changes to Sync Server.
       * Sync Server deploying confirmed changes to Subscribers.

    This class should enable lookup of basic room names regardless
    of the room name form.
    """

    def __getitem__(self, key: str) -> 'ConcordTriggerChannel':
        """
        Extract the base room name from the full room path.
        """
        key = trim_publisher_prefix(key)
        try:
            return super().__getitem__(key)
        except KeyError:
            return super().__getitem__(trim_tenant_cid(key, self))

    def __setitem__(self, name: str, value: str) -> None:
        super().__setitem__(name, value)

    def room_subscriptions(self):
        """
        Get all room names that this Appliance tenant should subscribe to.
        """
        rooms = []

        for base_room_name, channel in self.items():
            subscribe_room_name = channel.subscribe_room_name()
            rooms.append(subscribe_room_name)
            rooms.append(f'{subscribe_room_name}/#')

        return rooms


def trim_publisher_prefix(room_name: str) -> str:
    """
    Room name may be prefixed with Willow Publisher name and Appliance Tenant CID.

    willow.compass/Person
    willow.compass/Person/779233ef-ace8-4698-92bf-3d4088c527cd
    """
    publisher_room_prefix = f'{settings.PUBLISHER_ROOM_PREFIX}/'
    try:
        if room_name.startswith(publisher_room_prefix):
            return room_name[len(publisher_room_prefix) :]
    except AttributeError:
        pass

    return room_name


def trim_tenant_cid(room_name: str, registry: dict) -> str:
    """
    Room name may have Tenant CID suffix.

    Person/779233ef-ace8-4698-92bf-3d4088c527c
    """
    try:
        if room_name.startswith(tuple(registry.keys())):
            return room_name[: room_name.find('/')]
    except AttributeError:
        pass

    return room_name


def base_tenanted_record_publish_rooms(record: Model, room: str) -> list[str]:
    """
    Subscribers establish Room based on their Appliance CID,
    so that each subscriber can be targeted specifically.
    Records will not necessarily belong on all Appliances, and
    we want to avoid sending messages to a Room unless the record
    definitively does belong there.

    Sync Publishers must broadcast record changes to the various Topics
    to which Sync Subscribers may be subscribed. This function should
    calculate all the Topics that may be interested in this record.

    Provide custom functionality by setting
    CONCORD_MQ_TENANTED_PUBLISH_ROOMS="path.to.your.function"
    """
    if not settings.SYNC_PUBLISHER:
        # The provided room string should be correctly formed to
        # publish the message from local Subscriber to remote Publisher.
        return [room]

    return []



from concordia.models import (
    CanonicalIdDeleteLog,
    CanonicalIdMap,
    CanonicalIdSwap,
)
from .filters import (
    CanonicalIdDeleteLogFilter,
    CanonicalIdMapFilter,
    CanonicalIdSwapFilter,
)
from console_base.api.viewsets import LCModelViewSet
from .serializers import (
    CanonicalIdDeleteLogSerializer,
    CanonicalIdMapSerializer,
    CanonicalIdSwapSerializer,
)


# ---------------------------------------------------------------------------
class CanonicalIdDeleteLogViewSet(LCModelViewSet):
    queryset = CanonicalIdDeleteLog.objects.select_related('table')
    serializer_class = CanonicalIdDeleteLogSerializer
    filterset_class = CanonicalIdDeleteLogFilter


# ---------------------------------------------------------------------------
class CanonicalIdMapViewSet(LCModelViewSet):
    queryset = CanonicalIdMap.objects.select_related('table')
    serializer_class = CanonicalIdMapSerializer
    filterset_class = CanonicalIdMapFilter


# ---------------------------------------------------------------------------
class CanonicalIdSwapViewSet(LCModelViewSet):
    queryset = CanonicalIdSwap.objects.select_related('table')
    serializer_class = CanonicalIdSwapSerializer
    filterset_class = CanonicalIdSwapFilter
